# Copyright (C) 2020 Ludvig Larsson aka Valmond ludviglarsson@protonmail.com
# This code is licensed under the terms of the GNU Public License, version 3 or later (GPL-3.0-or-later)

import os
from Cryptodome.PublicKey import RSA
from encryption import CipherHelper
import json
import binascii
import re

import database_connection as dc
import rsa_helper as rh
import database_query as dq
from helpers import human_readable_file_size
from configuration import get_configuration_value, get_exclude_ip_list
from accept_swicharoo import do_accept_swicharoo
import locator

"""
Query helper for the database
"""


# Todo: include file sizes in the database instead
def get_shared_files_size():
    total_size = 0
    for dirpath, dirnames, filenames in os.walk("./data/"):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # skip if it is symbolic link or if it is our data
            if not os.path.islink(fp) and not fp[:12] == './data/local':
                #print(fp, os.path.getsize(fp))
                if os.path.exists(fp):
                    total_size += os.path.getsize(fp)
                else:
                    print("#Warning: database contains missing file:", fp)

    return total_size

# Todo: include file sizes in the database instead
def get_owned_files_size():
    total_size = 0
    for dirpath, dirnames, filenames in os.walk("./data/"):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # skip if it is symbolic link or if it is our data
            if not os.path.islink(fp) and fp[:12] == './data/local':
                #print(fp, os.path.getsize(fp))
                if os.path.exists(fp):
                    total_size += os.path.getsize(fp)
                else:
                    print("#Warning: database contains missing file:", fp)

    return total_size


def get_owned_file_size(dataid):
    my_name, _ = get_link_data_by_dataid(dataid)
    if my_name is not None:
        fp = 'data/local/' + my_name + '.aes'
        if os.path.exists(fp):
            return os.path.getsize(fp)
        else:
            print("#Warning: database contains missing file:", fp)
    return 0


# Returns true if IP is on the exclude list in the config file
def ip_on_exclude_list(ip):
    exclude_ip_list = get_exclude_ip_list()
    for exclude in exclude_ip_list:  # WAS several exclude regex, is for now just one.
        if re.search(exclude, ip):
            return True
    return False

# Returns true if this IP:PORT is my address
def is_my_ip_port(ip, port):
    if isinstance(port, int):
        port = str(port)
    my_port = get_configuration_value('port')
    if isinstance(my_port, int):
        my_port = str(my_port)
    return get_configuration_value('ip') == ip and my_port == port

# print(get_shared_files_size())
# print(get_owned_files_size())

#

# -------------------------- Links -----------------------------------

# --- The shared data (that we share for another node)


# Returns None if 'name'+'rsa_pub' does not have an entry in the database
# Else returns a row of NAME, ORIGINALFILENAME, KEY, IV, RSA_PUB
def get_link_data_by_name_and_rsapub(name, rsa_pub):  # Getting data for a link not owned by us, but stored by us (from a swicharoo)
    cursor = dc.execute("SELECT name, originalfilename FROM data_shared WHERE name = ? AND pubkey = ?", (name, rsa_pub))
    rows = cursor.fetchall()
    if len(rows) > 0:
        return rows[0]

    # So we doesn't share the data for anyone, maybe it's our original data someone wants?
    cursor = dc.execute("SELECT name, originalfilename FROM data_owned WHERE name = ? AND pubkey = ?", (name, rsa_pub))
    rows = cursor.fetchall()
    if len(rows) > 0:
        return rows[0]

    return None


# --- Our data that we would like other nodes to share

# Returns None if 'name' does not have an entry in the database
# Else returns a row of NAME, ORIGINALFILENAME, KEY, IV
def get_link_data_by_name(name):  # Getting data for a link owned by us, caller will get the RSA pub key from the database
    cursor = dc.execute("SELECT name, originalfilename, key, iv FROM data_owned WHERE name = ? ", (name,))  # comme to make the () a tuple...

    rows = cursor.fetchall()
    if len(rows) > 0:
        return rows[0]
    return None


# Returns None if 'name' does not have an entry in the database
# Else returns dataid
def get_link_dataid_by_name(name):  # Getting data for a link owned by us
    cursor = dc.execute("SELECT id FROM data_owned WHERE name = ? ", (name,))

    rows = cursor.fetchall()
    if len(rows) > 0:
        return rows[0][0]
    return None


# Returns None if 'name' does not have an entry in the database
# Else returns a row of NAME, ORIGINALFILENAME
def get_link_data_by_dataid(dataid):  # Getting data for a link owned by us, caller will get the RSA pub key from the database
    cursor = dc.execute("SELECT name, originalfilename FROM data_owned WHERE id = ? ", (dataid,))  # comme to make the () a tuple...

    rows = cursor.fetchall()
    if len(rows) > 0:
        return rows[0]
    return None, None


def get_random_local_data_id():  # todo: "WHERE wanted_fingers < SUM( ... fingers ...)"
    # return dc.query_one('SELECT id FROM data WHERE key != '' ORDER BY RANDOM() LIMIT 1;')
    # As it can return zero rows, we'll need to do it the old way
    cursor = dc.execute("SELECT id FROM data_owned WHERE key != '' ORDER BY RANDOM() LIMIT 1;")
    rows = cursor.fetchall()
    if len(rows) > 0:
        return rows[0][0]
    return None


def get_wanted_fingers(dataid):
    return dc.query_one('SELECT wanted_fingers FROM data_owned WHERE id = ?', (dataid,))


#

# -------------------------- Nodes -----------------------------------

#


# Takes any node and returns its ID, or None if no nodes
def get_random_node_id(only_validated_nodes=True):
    ip = get_configuration_value('ip')
    port = get_configuration_value('port')

    if only_validated_nodes:
        cursor = dc.execute('SELECT id FROM node WHERE last_connection_unixtime != 0 AND (ip != ? OR port != ?) ORDER BY RANDOM() LIMIT 1;', (ip, port))
    else:
        cursor = dc.execute('SELECT id FROM node WHERE ip != ? OR port != ? ORDER BY RANDOM() LIMIT 1;', (ip, port))
    rows = cursor.fetchall()
    if rows is None:
        return None
    if len(rows) > 0:
        return rows[0][0]
    return None


def get_node_data(nodeid):  # returns ip, port, RSAPubKey or None
    cursor = dc.execute("SELECT ip, port, pubkey FROM node WHERE id = ? LIMIT 1", (nodeid,))
    rows = cursor.fetchall()
    if len(rows) > 0:
        return rows[0]
    return None


def node_exists(ip, port, rsa_pub_key):  # ip:port no longer used to differentiate nodes
    cursor = dc.execute("SELECT id FROM node WHERE ip = ? and port = ? and pubkey = ?", (ip, port, rsa_pub_key, ))
    rows = cursor.fetchall()
    return len(rows) > 0


def get_nodeid(ip, port, rsa_pub_key):  # ip:port no longer used to differentiate nodes
    tmp = dc.query_one("SELECT id FROM node WHERE pubkey = ?", (rsa_pub_key, ))
    if tmp is None:
        return 0
    return tmp


def insert_node(ip, port, rsa_pub_key):
    if not node_exists(ip, port, rsa_pub_key):
        # Check it is not on the avoid list and not our IP:PORT
        if not ip_on_exclude_list(ip) and not is_my_ip_port(ip, port):
            print('Got new node, insert it: INSERT new node: ', ip, port, rsa_pub_key[-20:])
            dc.insert("INSERT INTO node (ip, port, pubkey) VALUES(?, ?, ?)", (ip, port, rsa_pub_key))
            #print('INSERT NODE rsa end:', rsa_pub_key[:12])


def update_node_quality(nodeid, success):
    connection = dc.get_connection()
    cursor = connection.cursor()
    if success:
        cursor.execute("UPDATE node SET connection_tries=connection_tries+1, last_connection_unixtime=DateTime('now') WHERE id = ? ", (nodeid,))  # comme to make the () a tuple...
    else:
        cursor.execute("UPDATE node SET connection_tries=connection_tries+1, connection_fails=connection_fails+1 WHERE id = ? ", (nodeid,))  # comme to make the () a tuple...
    connection.commit()


# returns the node id a finger points to
def get_node_from_finger(fingerid):
    tmp = dc.query_one("SELECT nodeid FROM finger WHERE id = ?", (fingerid, ))
    if tmp is None:
        return 0
    return tmp


# Deletes this data that we own, In The DataBase. Returns true if the data did exist, false otherwise
def delete_owned_data(owned_data_name):
    data_id = get_link_dataid_by_name(owned_data_name)
    if data_id:
        dc.update("DELETE FROM data_owned WHERE name = ?", (owned_data_name,))
        return True
    return False


#

# -------------------------- Fingers -----------------------------------

#

# How many fingers point on this data
def get_number_of_fingers(data_id):
    data_version = get_highest_version(data_id)
    tmp = dc.query_one("SELECT COUNT(*) FROM finger WHERE dataid = ? AND data_version = ?", (data_id, data_version))
    if tmp is None:
        return 0
    return tmp


# version or None if the finger does not exists
def get_finger_version(finger_id):
    return dc.query_one("SELECT data_version FROM finger WHERE id = ?", (finger_id, ))


# returns all fingerid pointing to this data
def get_all_fingers(data_id):
    rows = dc.query_all("SELECT id FROM finger WHERE dataid = ?", (data_id, ))
    r = []
    for id in rows:  # maybe this is not necessary... but as it should work, lets check that later
        r.append(id[0])
    return r


# returns all finger_id pointing to this data, with a specific version
def get_fingers_by_version(data_id, version):
    rows = dc.query_all("SELECT id FROM finger WHERE dataid = ? AND data_version = ?", (data_id, version))
    r = []
    for id in rows:  # maybe this is not necessary... but as it should work, lets check that later
        r.append(id[0])
    return r


# Runs through the fingers and returns the highest data_version, or 0 if none
def get_highest_version(data_id):
    fingers = dq.get_all_fingers(data_id)
    data_version = 0
    for finger_id in fingers:
        finger_version = dq.get_finger_version(finger_id)
        if finger_version > data_version:
            data_version = finger_version
    return data_version


# returns nodeid, dataid or 0, 0 if non existing
def get_finger_data(fingerid):
    cursor = dc.execute("SELECT nodeid, dataid FROM finger WHERE id = ?", (fingerid, ))
    rows = cursor.fetchall()
    if len(rows) > 0:
        return rows[0]
    return None
    

# Add one finger (e.g another node) that points on one of our data
# Also adds the correct address to this data
def insert_finger(nodeid, dataid):
    dc.insert("INSERT INTO finger (nodeid, dataid) VALUES (?, ?)", (nodeid, dataid))
    # Get ip, port, pubkey from the node
    node_info = get_node_data(nodeid)  # returns ip, port, RSAPubKey or None
    if node_info:
        # Add a digest
        ip = str(node_info[0])
        port = node_info[1]
        other_pub_key = str(node_info[2])
        # Get our private key so we can sign the address
        prv_hex, _ = dq.get_rsa_hex_keys()
        # Generate real key from the hex key
        prv = RSA.importKey(bytes.fromhex(prv_hex))
        # Get the public key from the node storing our data
        digest = rh.create_address_digest(prv, ip, port, other_pub_key)

        # Better even, encrypt the address with the symmetric cipher for the data
        # 1) get key & iv for this data
        tmp = get_link_data_by_dataid(dataid)
        if not tmp:
            print('#get_link_data_by_dataid error, dataid:', dataid)
            return
        name = tmp[0]
        row = dq.get_link_data_by_name(name)
        if not row:
            print('#get_link_data_by_name error, name:', name)
            return

        key = bytes.fromhex(row[2])
        iv = bytes.fromhex(row[3])

        # 2) make a json string with the address
        d = {'ip': ip, 'port': port, 'rsa_pub': other_pub_key}
        address_string = json.dumps(d)
        # print('insert_finger: address string =', address_string)

        # 3) encrypt
        ch = CipherHelper(aes_key=key, raw_iv=iv)
        encrypted_address = ch.encrypt(address_string)
        encrypted_address = binascii.hexlify(encrypted_address)
        encrypted_address = encrypted_address.decode()

        # 4) store it all (with digest for now)
        print('Store the address')
        add_address(dataid, ip, port, other_pub_key, digest, encrypted_address)

    else:
        print('')
        print('#Error insert_finger() cannot find nodeid')
        print('')


# returns fingerid or zero
def get_finger_id(nodeid, dataid):
    cursor = dc.execute("SELECT id FROM finger WHERE nodeid = ? AND dataid = ?", (nodeid, dataid))
    rows = cursor.fetchall()
    if len(rows) > 0:
        return rows[0]
    return None


def update_finger_quality(node_id, finger_id, success):
    connection = dc.get_connection()
    cursor = connection.cursor()
    if success:
        cursor.execute("UPDATE finger SET connection_tries=connection_tries+1, last_connection_unixtime=DateTime('now')  WHERE id = ? ", (finger_id,))  # comme to make the () a tuple...
    else:
        cursor.execute("UPDATE finger SET connection_tries=connection_tries+1, connection_fails=connection_fails+1  WHERE id = ? ", (finger_id,))  # comme to make the () a tuple...
    connection.commit()


# returns finger_id, or 0 if no finger found
def get_finger(my_data_id, other_node_id):
    r = dc.query_one("SELECT finger.id FROM finger JOIN node ON node.id = finger.nodeid WHERE finger.dataid = ? AND node.id = ? LIMIT 1", (my_data_id, other_node_id))
    if r:
        return r
    return 0

# Takes any finger and returns its ID, or None if there are no fingers in the database
def get_random_finger_id():
    return dc.query_one('SELECT id FROM finger ORDER BY RANDOM() LIMIT 1;')


def get_fingers_node_id(fingerid):
    return dc.query_one("SELECT nodeid FROM finger WHERE id = ?", (fingerid, ))


# When user updates data, other nodes will have stale data, get a finger pointing to a node in this situation
def get_stale_data_fingers():
    # Get all data we own (LIMIT a bit)
    stale_fingers = []
    # Get a bunch of data we share, to control:
    cursor = dc.execute("SELECT id, data_version FROM data_owned ORDER BY RANDOM() LIMIT 10")
    rows = cursor.fetchall()
    for row in rows:
        # check if the data is stale:
        dataid = row[0]
        data_version = row[1]
        cursor2 = dc.execute("SELECT id, nodeid, dataid, data_version FROM finger WHERE dataid = ? AND data_version < ?", (dataid, data_version))

        all_stale_fingers = cursor2.fetchall()
        for sf in all_stale_fingers:
            # Add the Actual, new data version:
            tmp = {0: sf[0], 1: sf[1], 2: sf[2], 3: sf[3], 4: data_version}
            stale_fingers.append(tmp)

    return stale_fingers


# Deletes all the fingers related to this data that we own. Returns true if the data exist, false otherwise.
def delete_fingers(owned_data_name):
    data_id = get_link_dataid_by_name(owned_data_name)
    if data_id:
        print('Delete fingers pointing on owned data id', data_id)
        dc.update("DELETE FROM finger WHERE dataid = ?", (data_id,))
        return True
    return False



#


############################################
#     Data

# Returns originalfilename or None if not existing
def get_originalfilename_from_data(name, rsa_pub_key):
    return dc.query_one("SELECT originalfilename FROM data_owned WHERE name = ? AND pubkey = ?", (name, rsa_pub_key))


# returns a dataid that doesn't yet have it's full number of fingers, or zero
# Do not use a data that is already stored by rsa_pub_key
# Checks if it is validated by our size policy, disk isn't full etc
def get_valid_dataid_needing_finger_for_swicharoo(other_file_size, other_rsa_pub_key, other_name):  # todo: use filesize in some way (+/- half magnitude, config, ...)
    other_node_id = get_nodeid('', 0, other_rsa_pub_key)

    for i in range(1, 50):  # todo: make this a SQL oneliner (join count fingers on data)
        rows = dc.query_all('SELECT id, name FROM data_owned ORDER BY RANDOM() LIMIT 1;')
        if len(rows) != 1:  # No files at all to share
            return 0
        row = rows[0]
        my_data_id = row[0]
        my_data_name = row[1]
        # Check other node isn't already storing it:
        if my_data_id is not None:
            # Check the other node doesn't already store this data:
            if get_finger(my_data_id, other_node_id) == 0:
                has_fingers = get_number_of_fingers(my_data_id)
                needs_fingers = get_wanted_fingers(my_data_id)
                if needs_fingers > has_fingers:

                    # Do we accept this file?
                    my_file_size = os.path.getsize(os.path.join(locator.create_path('data/local/'), (my_data_name + '.aes')))

                    if do_accept_swicharoo(my_file_size, other_file_size, logging=False):
                        return my_data_id

    return 0


# Tries to insert, there is a unique key for name+pubkey making duplicates not possible
def insert_data_shared(name, originalfilename, pubkey):
    dc.insert('INSERT INTO data_shared (name, originalfilename, pubkey) VALUES(?, ?, ?)', (name, originalfilename, pubkey))


# Updates the address list if it is newer. todo: later we should add unknown addresses too
def try_update_address_list(name, pub, address_list, epoch_time):
    old_epoc_time = dc.query_one('FROM data_owned SELECT address_list_epoc_time WHERE name = ? and pubkey = ? LIMIT 1', (name, pub))
    if old_epoc_time < epoch_time:  # update!
        print('Update address list for name:', name)
        dc.execute('UPDATE data_owned SET address_list = ?, address_list_epoc_time=?', (address_list, epoch_time))
    else:
        print('Abort insert, data is stale (too old)')


#

# -------------------------- Addresses -----------------------------------

############################################
#     Addresses for shared data

# Add a new address for this dataid
def add_address(dataid, ip, port, rsa_pub_key, digest, encrypted_address):
    # Check it is not on the avoid list and not our IP:PORT
    if not ip_on_exclude_list(ip) and not is_my_ip_port(ip, port):
        dc.insert('INSERT INTO address (pubkey, ip, port, dataid, digest, encrypted_address) VALUES(?, ?, ?, ?, ?, ?)', (rsa_pub_key, ip, port, dataid, digest, encrypted_address))


# Returns all the addresses pointing on this data
def get_encrypted_addresses(dataid):
    address_list = []
    cursor = dc.execute("SELECT encrypted_address FROM address WHERE dataid = ? ", (dataid, ))
    rows = cursor.fetchall()
    for row in rows:
        # print('address:', row[0])
        address_list.append(row[0])

    return address_list


# Returns all the addresses linked to this dataid
def get_address_list(nodeid, dataid):
    pass


#

# -------------------------- Owners Keys -----------------------------------

############################################
#     Owners RSA keys

# Returns two str
def get_rsa_hex_keys():  # Returns RSA keypair in hex format, or None if they do not exists / cannot be accessed.
    connection = dc.get_connection()
    cursor = connection.cursor()
    cursor.execute("SELECT privkey, pubkey FROM rsa")  # There should be only one entry here

    rows = cursor.fetchall()
    if len(rows) > 1:
        print('Warning! You have multiple RSA keypairs in your datbase! Returning first one...')
    if len(rows) > 0:
        row = rows[0]
        return row[0], row[1]
    return None, None


def set_rsa(prv, pub):
    connection = dc.get_connection()
    cursor = connection.cursor()
    textkey_prv = prv.exportKey('DER').hex()
    textkey_pub = pub.exportKey('DER').hex()
    print('Pubkey=', textkey_pub)

    cursor.execute("INSERT INTO rsa (privkey, pubkey ) VALUES(?, ?);", (textkey_prv, textkey_pub))
    connection.commit()

#

# -------------------------- Folders -----------------------------------

#


# Returns stored folder, or inserts new if there are no one
def get_folder(pub_key):
    _, my_pub = get_rsa_hex_keys()
    if my_pub == pub_key:
        return 'local'
    tmp = dc.query_one("SELECT folderid FROM folder WHERE pubkey = ? ", (pub_key, ))
    if tmp:
        return str(tmp)

    dc.insert("INSERT INTO folder (pubkey) VALUES(?);", (pub_key, ))

    tmp = dc.query_one("SELECT folderid FROM folder WHERE pubkey = ? ", (pub_key,))
    if tmp is None:
        return 'error'
    return str(tmp)

#

# -------------------------- Flush information to console -----------------------------------

#


# prints information about our own owned data
def flush_owned_data():
    rows = dc.query_all("SELECT id, name, originalfilename, data_version FROM data_owned")
    print('Owned data:', len(rows))
    print('<fingers> <Name>    <version>   <Original file/folder name>   <size> ')
    data_owned = []
    for data in rows:  # maybe this is not necessary... but as it should work, lets check that later
        data_id = data[0]
        data_version = data[3]
        wanted_fingers = get_wanted_fingers(data_id)
        actual_fingers = get_number_of_fingers(data_id)
        file_size = get_owned_file_size(data_id)
        print(str(actual_fingers)+'/'+str(wanted_fingers) + ' \t  ['+str(data[1])+']  \t' + str(data_version)+ ' \t '+str(data[2]), ' \t ' + human_readable_file_size(file_size))


# prints information about shared data
def flush_shared_data():
    rows = dc.query_all("SELECT id FROM data_shared")
    print('Number of shared data:', len(rows), '  Total size: ' + human_readable_file_size(get_shared_files_size()))

# Shows the success ratio for all nodes with success>0
# Also shows the number of nodes at zero success
def flush_known_nodes():
    bad_or_new_nodes = dc.query_one("SELECT COUNT (*) FROM node WHERE connection_tries = connection_fails AND connection_tries > 0")
    never_connected_to_nodes = dc.query_one("SELECT COUNT (*) FROM node WHERE connection_tries = 0")
    rows = dc.query_all("SELECT ip, port, connection_tries, connection_fails FROM node WHERE connection_tries > connection_fails AND connection_tries > 0")

    print('Good nodes: ' + str(len(rows)) + '  (If any, they will be listed below)')
    print('New or bad nodes: ' + str(bad_or_new_nodes) + '  (We have tried but never achieved a connection to these nodes)')
    print('Never connected to nodes: ' + str(never_connected_to_nodes) + '  (Debug nodes, not yet checked out nodes...)')

    print('')
    print('Good nodes:')
    for data in rows:
        ip = data[0]
        port = data[1]
        tries = data[2]
        fails = data[3]
        success = tries - fails
        success_rate = int((10000 * success)/tries) / 100
        print('Node: address=' + ip + ':' + port + ' tries/fails: '+str(tries)+'/'+str(fails)+ ' ['+str(success_rate)+'% success]' )
    # count of bad nodes/too new nodes to have a ratio:






# Shows information of one of our owned data
def flush_fingers(name):
    dataid = get_link_dataid_by_name(name)
    if dataid:
        rows = dc.query_all("SELECT id, nodeid, data_version, connection_tries, connection_fails, last_connection_unixtime FROM finger WHERE dataid = ?", (dataid,))
        r = []
        print('name id  nodeid  version   tries/fails [%]   <last successful finger connection>')
        for data in rows:  # maybe this is not necessary... but as it should work, lets check that later
            percent = '---'
            tries = data[3]
            fails = data[4]
            success = tries - fails
            if tries > 0:
                percent = str(int(100 * success / tries))+'%'
            print(name, ':',  data[0], data[1], data[2], str(data[3]) +'/' + str(data[4]), percent, str(data[5]))



    else:
        print('Unknown data:', name)

