# Copyright (C) 2020 Ludvig Larsson aka Valmond ludviglarsson@protonmail.com
# This code is licensed under the terms of the GNU Public License, version 3 or later (GPL-3.0-or-later)

import os
import sys
import time
import json

from struct import pack, unpack
from Cryptodome.PublicKey import RSA

from socket_communication import SecureCommunication
import database_query as dq
import updater as ud
import rsa_helper as rh
import database_connection as dc
from configuration import get_configuration_value


# todo: randomly call requre_update_data_information for fingers


"""
The scheduler makes certain commands at certain intervals like:

* Trying to find new nodes
* Check if nodes are up (and update the DB accordingly)
* Remove stale nodes
* Check that all our data have 10 fingers, or try to get a new finger
* Verify fingers pointing on our data (and update the DB accordingly)
* Remove stale fingers

"""

"""
For convenience, we'll make the different functions in this file
Later on when there are several, we should move them to separate source files.

"""

_log_event_calls = False  # prints to stdout all the calls the scheduler does

###################################
#       DB helpers
#

# Updates the nodes quality
def update_node_quality(node_id, success):
    # Updates db::node: CONNECTION_TRIES, CONNECTION_FAILS, LAST_CONNECTION_UNIXTIME
    # get a random node:
    node_id = dq.get_random_node_id(only_validated_nodes=False)
    if node_id is not None:
        info = dq.get_node_data(node_id)
        sc = SecureCommunication(info[0], info[1], info[2]) #  IT IS NOT str(...) but ENCODE/DECODE!!!!!
        dq.update_node_quality(node_id, sc.is_active())
        sc.close()


# Does not update the nodes quality, only the fingers quality
# MOVED TO database_query.py
def _DEPRECATED_update_finger_quality(finger_id, success):
    # Updates db::finger: CONNECTION_TRIES, CONNECTION_FAILS, LAST_CONNECTION_UNIXTIME
    finger_id = dq.get_random_finger_id()
    if finger_id is not None:
        node_id = dq.get_node_from_finger(finger_id)
        info = dq.get_node_data(node_id)
        sc = SecureCommunication(info[0], info[1], info[2])
        dq.update_finger_quality(node_id, sc.is_active())
        sc.close()

#


###################################
#       NODES
#


# Returns a secure, functionnal node, or None
def get_random_node_connection(only_validated_nodes=True):
    node_id = dq.get_random_node_id(only_validated_nodes)  # todo: remove the False in the call
    if node_id is None:
        return None, 0
    tmp = dq.get_node_data(node_id)
    ip = str(tmp[0])
    port = tmp[1]
    rsa_pub_key = str(tmp[2])
    # print('Try to connect to random node:', ip, port) #, rsa_pub_key)
    # print('using rsa type, len', type(rsa_pub_key), len(rsa_pub_key))
    # print('using rsa', rsa_pub_key[-10:])
    sc = SecureCommunication(ip, port, RSA.importKey(bytes.fromhex(rsa_pub_key)))  # rsa_pub here is the one linked to the other Node
    if sc.is_active():
        return sc, node_id
    return None, node_id


# One of the scheduled events:
def find_new_nodes():
    # Get a random nod in our db:node, connect to it and ask for some nodes
    sc, node_id = get_random_node_connection(False)  # todo: remove the False in the call

    if sc is None:
        #print('Random node is not up (Find new Nodes).')
        return
    #print('Ask for nodes from node', node_id)
    # Query for some random node(s)
    ip = get_configuration_value('ip')
    port = get_configuration_value('port')
    _, rsa_pub = rh.get_rsa_keypair(True)

    # Send which port we prefer
    port = str(port).encode()
    # and our public key
    rsa_pub = rsa_pub.exportKey('DER').hex().encode()

    #print('Send SEND_NODES req.', flush=True)
    json_answer = sc.send(b'SEND_NODES' + port + b',' + rsa_pub) # ,' + aa + b',' + bb + b',' + rr)  # expected return: json with a node: ip, port, rsa_pub_key, or an empty json if no nodes: {}
    #print('Send SEND_NODES req. Got Answer:', json_answer[:48], flush=True)
    if json_answer:
        try:
            #print('Got an answer for our SEND NODES, lets try to json decode it:')
            j = json.loads(json_answer.decode())  # decode to string and then put JSON-data to a variable
            # insert it in the database
            if 'ip' in j and 'port' in j and 'rsa_pub_key' in j:
                # convert ip & rsa_pub_key from bytes to str, port is an integer so no need.
                #print('Got node:', str(j['ip']), ':', j['port'])
                dq.insert_node(str(j['ip']), j['port'], str(j['rsa_pub_key']))
            else:
                print('#No information in json, bailing out:', json_answer.decode())
        except:
            print('JSON Exception!')
            print('tried to decode:', json_answer.decode())
            return
    else:
        print('Got empty answer. Bail out.')


# don't forget to use update_node_quality()


# One of the scheduled events:
def check_node_quality():
    # update connection_tries & connection_fails
    # update last_connection_datetime if the connection worked
    sc, node_id = get_random_node_connection(False)  # todo: remove the False in the call
    node_up = False
    if sc is not None:
        node_up = sc.is_active()

    # print('### UPDATE NODE QUALITY: ', node_id, node_up)
    dq.update_node_quality(node_id, node_up)

    if node_up:
        sc.close()


# One of the scheduled events:
def remove_stale_nodes():
    # When we have a large number of nodes, cull the ones no longer working
    # by removing the oldest db:nodes::last_connection_datetime
    pass

#

###################################
#       Our Data
#


def _update_finger_data(finger_id, node_id, data_id, new_data_version, file_size):
    print('Try to update finger id:', finger_id)
    tmp = dq.get_node_data(node_id)
    ip = str(tmp[0])
    port = tmp[1]
    other_rsa_pub_key = str(tmp[2])
    sc = SecureCommunication(ip, port, RSA.importKey(bytes.fromhex(other_rsa_pub_key)))
    if not sc.is_active():
        # node is not up
        print('Could not connect for update, bail out.')
        return False

    # get data information & data
    my_name, my_original_file_name = dq.get_link_data_by_dataid(data_id)

    # get our PubKey:
    prv, pub = rh.get_rsa_keypair()
    _, rsa_pub_hex = dq.get_rsa_hex_keys()
    #rsa_pub_hex = pub.exportKey('DER').hex().encode()

    # ask for an update
    query = b'UPDATEDATA' + pack('>Q', file_size) + b',' + my_name.encode() + b',' + rsa_pub_hex.encode() + b','

    random_checksum = sc.send(query)
    # make a digest and send back for authentification, prove we actually own the data to be updated:
    digest = rh.create_digest(random_checksum, prv)
    print('Calculating digest:', digest)
    # send back digest
    answer = sc.send(digest)
    if answer[:10] != b'ACCEPTING!':
        print('Refresh data query was turned down:', answer[:10])
        sc.close()
        return False
    # Refresh accepted, lets go through with it:

    ip = get_configuration_value('ip')
    # get our PubKey:
    if not prv:
        return False

    with open('data/local/' + my_name + '.aes', 'rb') as f:
        file_data = f.read()
    if file_data is None:
        print('Refresh: File not found', my_original_file_name)
        sc.close()
        return False

    other_ip, other_port, other_pub_key = dq.get_node_data(node_id)
    answer = sc.send(file_data)
    if answer != b'OK':
        print('Refresh sending data failed')
        return False

    sc.close()
    print('Refresh successful!')
    print("UPDATE finger SET data_version = ? WHERE id = ?", (new_data_version, finger_id))
    dc.update("UPDATE finger SET data_version = ? WHERE id = ?", (new_data_version, finger_id))


# Try to get a new swicharoo for one of our data. Private function used by check_data_fingers()
# Returns true on success, otherwise False
def _get_new_finger(data_id):
    # As socket communication can fail in so many ways and we even know we'll be having stale addresses in our database, make a big try-catch here:
    sc = None
    try:
        # Connect to a random node
        # todo: only select nodes that do not store our data already!
        sc, node_id = get_random_node_connection(False)  # todo: remove the False in the call (use only verified nodes)
        if sc is None:
            return False

        # get our PubKey, ip, port
        prv, pub = dq.get_rsa_hex_keys()
        if not prv:
            sc.close()
            return False
        ip = get_configuration_value('ip')
        port = get_configuration_value('port')

        # get data information
        my_name, my_original_file_name = dq.get_link_data_by_dataid(data_id)

        # print('Check we do not already have the finger (node_id, data_id)', node_id, data_id)
        # Verify we are not already have a finger with this node:
        # todo: don't ask nodes that already store our data, this seems it isn't enough, maybe as sometimes we accidentally maskerade with different ip:ports
        # Achtually it seems like we check IP:PORT and not the actual RSA Pubkey?
        # Or just we contact the same node, but over another IP (like a 192. ... instead of public IP) we think they are different
        # but they are actually the same => they already store our data
        finger_id = dq.get_finger_id(node_id, data_id)
        # print('Scheduler: We got a finger_id:', finger_id)
        if finger_id is not None:  # This node already stores the data
            # print('Scheduler: We already have data stored at this node')
            sc.close()
            return False

        # print('Scheduler: Ask for new finger for data_id='+str(data_id)+' node_id='+str(node_id)+ ' data name='+my_name+' (origname='+my_original_file_name+')')
        # get the actual datasize from the data_id
        file_size = dq.get_owned_file_size(data_id)

        # ask for a swicharoo
        query = b'SWICHAROO?' + pack('>Q', file_size) + pub.encode() + b'|' + my_name.encode() + b'|' + ip.encode() + b'|' + str(port).encode()
        #   <<        Send Recv           >>
        answer = sc.send(query)
        if answer[:10] != b'YES_PLEASE':
            if answer[:10] == b'ALREADY_DO':
                # They already store our data, duh
                print('Scheduler: SWICHAROO query was turned down because they already store our data, node_id, data_id:', node_id, data_id)
                pass
            else:
                if answer[:10] == b'SRYNODATA.':
                    print('They turned down because they have nothing to share.')
                else:
                    print('Scheduler: SWICHAROO query was turned down, reason:', answer)
            sc.close()  # Close as this node does not behave well right now
            return False

        # Get size other node want us to store and check if we are OK:
        data = answer[10:]  # Remove the b'YES_PLEASE'
        data_size = data[:8]
        data = data[8:]  # Remove the size
        other_file_size = unpack('>Q',  data_size)[0]
        print('other_file_size:', other_file_size, flush=True)

        if other_file_size > get_configuration_value('swicharoo_file_maxsize_kb')*1024:
            sc.send(b'U_TOO_BIG!')
            sc.close()  # Close as the node has no more space anyways
            print('Scheduler: their file is too big (max=' + str(get_configuration_value('swicharoo_file_maxsize_kb')*1024) + '): their file size=', other_file_size)
            return False

        split = data.split(b'|')
        if len(split) != 3:
            sc.send(b'ERROR_LEN!')
            print('Split should be 3, is:', len(split))
            sc.close()  # Close as the node has no more space anyways
            return False

        # Get all information from the other node
        other_name = split[0].decode()
        other_original_file_name = split[1].decode()
        print('other_name:', other_name)
        digest_query = split[2]  # probably do not decode() this but use as is

        if digest_query is None or not sc.is_active():
            print('Other node refused our swicharoo query, bailing out too.', flush=True)
            sc.close()
            return False

        # If we should do it this way, we should also discern nodes by rsa,ip,port? or have several ip:port for 1 node...
        other_ip, other_port, other_pub_key = dq.get_node_data(node_id)  # Why do we both accept these data from the listener, but also get them from our database?

        # Check if we already store the data:
        if dq.get_link_data_by_name_and_rsapub(other_name, other_pub_key):  # We already store this data for this node!
            print('Scheduler: our swicharoo failed as we already store the data the other node proposes to switch.')
            sc.send(b'DATA_AGAIN')
            sc.close()
            return False

        print('Do digest, digest_query=', digest_query, flush=True)
        print('Do digest type=', type(digest_query), flush=True)

        digest = rh.create_digest(digest_query, RSA.importKey(bytes.fromhex(prv)))
        print('Okay all systems go!', flush=True)
        #   <<        Send Recv           >>
        other_data = sc.send(digest)

        # Save their data
        file_folder = 'data/' + dq.get_folder(other_pub_key) + '/'
        if not os.path.exists(file_folder):
            os.makedirs(file_folder)
        save_name = file_folder + other_name + '.aes'
        print('Scheduler: Swicharoo, save off others file:', save_name, flush=True)
        with open(save_name, 'wb') as f:
            f.write(other_data)

        # I now have other: pubkey, name, origname, data, save off data and update DB
        dq.insert_data_shared(other_name, other_original_file_name, other_pub_key)  # Others data is now stored here for retrieval (by name + pubkey)

        print('Got their data, send ours:', flush=True)

        # send our data, expect b'ok' back
        # get the data
        file_data = None
        with open('data/local/' + my_name + '.aes', 'rb') as f:
            file_data = f.read()
        if file_data is None:
            print('Scheduler: Our original file not found:', my_original_file_name, flush=True)
            sc.close()
            return False

        # Send our data
        #   <<        Send Recv           >>
        ack = sc.send(file_data)
        print('Got final ACK:', ack, flush=True)

        # Insert new finger for my data
        dq.insert_finger(node_id, data_id)
        print("Scheduler: The Swicharoo was a success; inserted a finger. NodeId, DataId:", node_id, data_id, flush=True)

        sc.close()
        return True

    except:
        if sc is not None:
            sc.close()
    return False


# One of the scheduled events:
def check_data_fingers():  # Checks all of our data has 10 fingers, or else try to get a new swicharoo
    data_id = dq.get_random_local_data_id()
    if data_id is not None:
        # Try to get 2 new fingers for a random data (if needed)
        check_data_finger(data_id, 2)


# Tries 'tries' times to get a new finger for this data, if needed
def check_data_finger(data_id, tries=10, print_output=False):
    newly_acquired_fingers = 0
    fingers = dq.get_number_of_fingers(data_id)
    wanted_fingers = dq.get_wanted_fingers(data_id)
    if print_output:
        if fingers >= wanted_fingers:
            print('This data already have a maximum number of fingers (' + str(fingers) + ')')
        else:
            print('Data has ' + str(fingers) + ' of ' + str(wanted_fingers) + ' fingers, try ' + str(tries) + ' times to get new ones.', flush=True)
    while fingers < wanted_fingers and tries > 0:
        if _get_new_finger(data_id):
            fingers += 1
            newly_acquired_fingers += 1
            if print_output:
                print('New finger acquired (' + str(fingers) + '/' + str(wanted_fingers) + ')', flush=True)
        tries = tries - 1
    if print_output:
        print('Newly acquired fingers:', newly_acquired_fingers)

# One of the scheduled events:
def verify_finger():  # Take one random finger and check if it's responding, update DB accordingly
    # take a random finger and call up it's node
    # don't forget to use update_node_quality() AND update_finger_quality()
    finger_id = dq.get_random_finger_id()
    if finger_id is not None:  ###### THE BUG IS ERE SOMEWHERE (inserting ip, rsa as byte, not string!), but actually it's just showing up here... :-/
        # Get its node:
        #print('verify_finger', finger_id)
        node_id = dq.get_fingers_node_id(finger_id)
        #print('verify_finger node:', node_id)
        if node_id is not None:
            tmp = dq.get_node_data(node_id)
            if tmp:
                ip = str(tmp[0])
                port = tmp[1]
                rsa_pub_key = str(tmp[2])  # Was this the bug??! Check and keep only this if so
                sc = SecureCommunication(ip, port, RSA.importKey(bytes.fromhex(rsa_pub_key)))  # rsa_pub here is the one linked to the other Node
                dq.update_finger_quality(node_id, finger_id, sc.is_active())
                sc.close()
            else:
                print('Warning: finger', finger_id, 'points on non existing node:', node_id)


# One of the scheduled events:
def remove_stale_fingers():  # Sometimes we'll have to let go of a finger because it doesn't work well enough
    # For a random data that has 10 fingers
    # remove the oldest db:nodes::last_connection_datetime, if older than some day

    pass


# Here we check if user has updated data we own, and that other nodes share
# If so, lets try to send the new data to the nodes
def refresh_stale_fingers():
    stale_fingers_data = dq.get_stale_data_fingers()
    for finger_data in stale_fingers_data:
        finger_id = finger_data[0]
        node_id = finger_data[1]
        data_id = finger_data[2]
        data_version = finger_data[3]  # The old version
        new_data_version = finger_data[4]

        print('STALE FINGER id=', finger_id)
        # Ask for a refresh
        file_size = dq.get_owned_file_size(data_id)
        _update_finger_data(finger_id, node_id, data_id, new_data_version, file_size)


#

#

###################################
#       The actual Scheduling mechanism
#

def run_scheduler(run_continously=True):
    wait_time = 1.0
    first_run = True
    while run_continously or first_run:
        first_run = False

        # Nodes
        if _log_event_calls:
            print('-------------------------------------------------------- 1')
            print('[Nodes]#Find new nodes:')
        find_new_nodes()
        time.sleep(wait_time)  # Don't use all the CPU/Bandwith

        if _log_event_calls:
            print('-------------------------------------------------------- 2')
            print('[Nodes]#Check node quality')
        check_node_quality()
        time.sleep(wait_time)  # Don't use all the CPU/Bandwith

        if _log_event_calls:
            print('-------------------------------------------------------- 3')
            print('[Nodes]#Prune stale nodes')
        remove_stale_nodes()
        time.sleep(wait_time)  # Don't use all the CPU/Bandwith

        # Data
        if _log_event_calls:
            print('******************************************************** 4')
            print('[Data]#Check data fingers')
        check_data_fingers()
        time.sleep(wait_time)  # Don't use all the CPU/Bandwith

        if _log_event_calls:
            print('******************************************************** 5')
            print('[Data]#Verify fingers')
        verify_finger()
        time.sleep(wait_time)  # Don't use all the CPU/Bandwith

        if _log_event_calls:
            print('******************************************************** 6')
            print('[Data]#Prune stale fingers')
        remove_stale_fingers()
        time.sleep(wait_time)  # Don't use all the CPU/Bandwith

        if _log_event_calls:
            print('******************************************************** 7')
            print('[Data]#Refresh stale owned data in nodes sharing older versions')
        refresh_stale_fingers()
        time.sleep(wait_time)  # Don't use all the CPU/Bandwith


class Scheduler:
    def __init__(self):
        print("Starting scheduler:")
        self.continue_to_run = True
        self.wait_time = 1.0

    def forcestop(self):
        print("Shutting down scheduler...")
        self.continue_to_run = False

    def run(self):
        while self.continue_to_run:
            # We're in a thread here (hence the class) so just run once, then see if we should shut down or run again
            run_scheduler(run_continously=False)
            time.sleep(self.wait_time)


def test():
    for i in range(1, 9):
        dq.update_node_quality(1, True)
    for i in range(1, 5):
        dq.update_node_quality(1, False)



# This will only happen if you run the file, not if you import it
if __name__ == '__main__':
    if len(sys.argv) > 1 and sys.argv[1] == '--testrun':
        print('[scheduler.py] Testrun')
        test()
        sys.exit(0)

    if len(sys.argv) > 1 and sys.argv[1] == '--multitestrun':
        print('[scheduler.py] Multitestrun')
        _log_event_calls = True
        run_scheduler()
    else:
        run_scheduler()

